from typing import List, Optional

from appium.webdriver.appium_service import AppiumService
from behave.configuration import Configuration as BehaveConfiguration
from behave.model import Feature, Tag, Scenario, Table, Row
from selenium.webdriver.remote.webdriver import WebDriver
from rikki.mitmproxy.plugin import ProxyPlugin


class Configuration(BehaveConfiguration):

    def __init__(self, command_args=None, load_config=True, verbose=None, **kwargs):
        super().__init__(command_args, load_config, verbose, **kwargs)
        self.proxy: Optional[ProxyPlugin] = None


class Context:
    """
    Provide typed access to Behave context object.
    More documentation can'be founded there https://behave.readthedocs.io/en/latest/context_attributes.html
    """

    def __init__(
            self,
            config: Optional[Configuration] = None,
            aborted: Optional[bool] = None,
            failed: Optional[bool] = None,
            feature: Optional[Feature] = None,
            tags: Optional[List[Tag]] = None,
            active_outline: Optional[Row] = None,
            scenario: Optional[Scenario] = None,
            table: Optional[Table] = None,
            text: Optional[str] = None,
            proxy: Optional[ProxyPlugin] = None,
            browser: Optional[WebDriver] = None,
            appium: Optional[AppiumService] = None,
    ) -> None:
        super().__init__()
        self.config: Optional[Configuration] = config
        self.aborted: Optional[bool] = aborted
        self.failed: Optional[bool] = failed
        self.feature: Optional[Feature] = feature
        self.tags: Optional[List[Tag]] = tags
        self.active_outline: Optional[Row] = active_outline
        self.scenario: Optional[Scenario] = scenario
        self.table: Optional[Table] = table
        self.text: Optional[str] = text
        self.proxy: Optional[ProxyPlugin] = proxy
        self.browser: Optional[WebDriver] = browser
        self.appium: Optional[AppiumService] = appium



import threading

from typing import Optional, List

from mitmproxy import http

from rikki.mitmproxy.model import Request, Response
from rikki.mitmproxy.plugin import ProxyDelegate
from rikki.mitmproxy.util import filter


class DelayConfig:

    def __init__(
            self,
            request: Optional[Request] = None,
            response: Optional[Response] = None,
            delay: int = 0
    ) -> None:
        """
        :param request: request data for filter
        :param response: response data for filter
        :param delay:
        """
        self.request = request
        self.response = response
        self.delay = delay


class Delay(ProxyDelegate):
    """
    This plugin will delay request processing based based on provided config.

    """

    def __init__(self, config: DelayConfig):
        super().__init__()
        self.__config: DelayConfig = config

    def response(self, flow: http.HTTPFlow):
        self.__delay_flows(filter([flow], request=self.__config.request, response=self.__config.response))

    def __delay_flows(self, flows: List[http.HTTPFlow]):
        for flow in flows:
            flow.intercept()
        threading.Timer(self.__config.delay, self.__resume_flows, args=[flows]).start()

    @staticmethod
    def __resume_flows(flows: List[http.HTTPFlow]):
        for flow in flows:
            flow.resume()



from typing import Dict, Optional

from mitmproxy import http
from mitmproxy.test import tflow as flow_creator

from rikki.mitmproxy.model import Request, Response


def build_flow(
        request: Optional[Request] = None,
        response: Optional[Response] = None
) -> http.HTTPFlow:
    flow: http.HTTPFlow = flow_creator.tflow(resp=response is not None)

    if request:
        flow.request.host = request.host

        if request.headers:
            for header in request.headers.items():
                flow.request.headers[header[0]] = header[1]

        if request.params:
            for param in request.params.items():
                flow.request.query.set_all(param[0], [param[1]])

    if response:
        if response.code:
            flow.response.status_code = response.code
        if response.headers:
            for header in response.headers.items():
                flow.response.headers.set_all(header[0], [header[1]])
        if response.content:
            flow.response.content = response.content.encode("utf8")

    return flow



from unittest import TestCase

from rikki.mitmproxy.delegate_accumulate import AccumulateFilterConfig, Accumulate
from rikki.mitmproxy.model import Request, Response
from rikki.mitmproxy.util import filter
from test.rikki.mitmproxy.utils import build_flow


class TestAccumulateDelegate(TestCase):

    def test_accumulate_by_request(self):
        expected_request = Request(host="required_host")
        flow_one = build_flow(request=expected_request)
        flow_two = build_flow(request=Request(host="not_required_host"))

        config = AccumulateFilterConfig(request=expected_request)

        instance = Accumulate(filter_config=config)
        instance.request(flow_one)
        instance.request(flow_two)

        assert len(instance.accumulated_items) == 1
        assert len(filter(instance.accumulated_items, request=expected_request))

    def test_accumulate_by_response(self):
        expected_response = Response(content="required response")
        flow_one = build_flow(response=expected_response)
        flow_two = build_flow(response=Response(content="not required response"))

        config = AccumulateFilterConfig(response=expected_response)

        instance = Accumulate(filter_config=config)
        instance.response(flow_one)
        instance.response(flow_two)

        assert len(instance.accumulated_items) == 1
        assert len(filter(instance.accumulated_items, response=expected_response))

    def test_accumulate_by_request_and_response(self):
        expected_request = Request(host="required_host")
        expected_response = Response(content="required response")
        flow_one = build_flow(request=expected_request, response=expected_response)
        flow_two = build_flow(
            request=Request(host="not_required_host"),
            response=Response(content="not required response")
        )

        config = AccumulateFilterConfig(
            request=expected_request,
            response=expected_response
        )

        instance = Accumulate(filter_config=config)
        instance.request(flow_one)
        instance.response(flow_one)
        instance.request(flow_two)
        instance.response(flow_two)

        assert len(instance.accumulated_items) == 1
        assert len(filter(instance.accumulated_items, request=expected_request, response=expected_response))



from typing import Optional

from mitmproxy import http

from rikki.mitmproxy.model import Request, Response
from rikki.mitmproxy.plugin import ProxyDelegate
from rikki.mitmproxy.util import filter as filter_flows, replace


class ReplacementConfig:

    def __init__(
            self,
            filter_request: Optional[Request] = None,
            filter_response: Optional[Response] = None,
            replacement_request: Optional[Request] = None,
            replacement_response: Optional[Response] = None,
            inclusive=False
    ) -> None:
        """
        This config represents data to be used in @InterceptAndReplace class.
        :param filter_request:
        :param filter_response:
        :param replacement_request: Will be used to replace data in request if `request_filter` presented
        :param replacement_response: Will be used to replace data in response if `request_filter` or `response_filter`
        :param inclusive: when True will replace all data for Request/Response, if not will add the data to them and
        replace only conflicted properties
        presented.
        """
        super().__init__()
        self.filter_request: Optional[Request] = filter_request
        self.filter_response: Optional[Response] = filter_response
        self.replacement_request: Optional[Request] = replacement_request
        self.replacement_response: Optional[Response] = replacement_response
        self.inclusive = inclusive


class InterceptAndReplace(ProxyDelegate):
    """
    This delegates responsible for intercepting and replacing of data in Request or Response
    """

    def __init__(self, config: ReplacementConfig) -> None:
        """
        Initialise interception and replace plugin. The plugin will intercept requests or response
        :param request:
        :param response:
        """
        super().__init__()
        self.__config = config

    def request(self, flow: http.HTTPFlow):

        filtered = filter_flows([flow], request=self.__config.filter_request)
        for filtered_flow in filtered:
            if self.__config.replacement_request:
                self.__prepare_request(filtered_flow)
                replace(flow=filtered_flow, request=self.__config.replacement_request)

    def response(self, flow: http.HTTPFlow):

        filter_request = self.__config.filter_request

        if self.__config.replacement_request:
            filter_request = self.__config.replacement_request

        filtered = filter_flows([flow], request=filter_request,
                                response=self.__config.filter_response)
        for filtered_flow in filtered:
            if self.__config.replacement_response:
                self.__prepare_response(filtered_flow)
                replace(flow=filtered_flow, response=self.__config.replacement_response)

    def __prepare_request(self, flow: http.HTTPFlow):
        if self.__config.inclusive:
            if self.__config.replacement_request:
                if self.__config.replacement_request.headers:
                    flow.request.headers.clear()
                if self.__config.replacement_request.params:
                    flow.request.query.clear()

    def __prepare_response(self, flow: http.HTTPFlow):
        if self.__config.inclusive:
            if self.__config.replacement_response:
                if self.__config.replacement_response.headers:
                    flow.response.headers.clear()
